function  [A]= conv_diff(n1,n2,n3,n4,n5,potent,alp1,alp2,alp3,alp4,alp5)
%
% Usage: [A]=conv_diff(n1,n2,n3,n4,n5,potent,alp1,alp2,alp3,alp4,alp5);
%    or: [A]=conv_diff(n1);
%        [A]=conv_diff(n1, n2);
%        [A]=conv_diff(n1, n2, n3);
%        [A]=conv_diff(n1, n2, n3, n4);
%        [A]=conv_diff(n1, n2, n3, n4, n5);
%
%  conv_diff generates 3-, 5-, 7-, 9- or 11- point finite difference 
%  discretization for  1d, 2d, 3d, 4d, or 5d convection-diffusion 
%  equation on the unit domain with zero Dirichlet boundary 
%  conditions. It is very easy to get very large dimension test 
%  matrices with this script.
%
%  the alp1,alp2,alp3,alp4,alp5,potent are all optional, 
%  they are set to zeros if not provided in the input list.
%
%  n2, n3, n4 and n5 are also optional.  
%  if only one integer is provided, it outputs a 1-d Laplacian,
%  if two integers are provided, it outputs a 2-d Laplacian,
%  if three integers are provided, it outputs a 3-d Laplacian.
%  this rule goes upto 5-d Laplacian in this script 
%  (theoretically one can go up any dimension as one wants).
%
%  if alpi=0 (i=1:5) and potent= 0, then the returned A is:
%     1d  Laplacian  if  only one of the (n1, n2, n3, n4, n5) > 1
%     2d  Laplacian  if  two of the (n1, n2, n3, n4, n5) > 1
%     i.e., the semi-physical dimension of the Laplacian is 
%     sum([n1, n2, n3, n4, n5]>1);
%     and the corresponding matrix dimension is n1*n2*n3*n4*n5.
%
% Note that if the output is a Laplacian, then some of the dimensions 
% (n1 to n5) can be omitted or input in any order.
% But if one wants to output a Hamiltonian, (i.e., the "potent" 
% shift is nonzero), then all the first five (n1 to n5) have to be 
% listed as input, and  the value of "potent" should be listed as the 6th 
% input variable.
% E.g., both of the following
%   conv_diff(10, 1, 1, 20, 30, -5) 
%   conv_diff(10, 20, 30, 1, 1, rand(6000,1)) 
% return a 3-d Hamiltonian. Note that if "potent" is not a scalar shift,
% its dimension has to agree with the actual matrix dimension.
%
%
%-- y.k. zhou  (zhou@msi.umn.edu)
%   March, 2005
%
  
  if (~exist('n1', 'var'))
    n1 = 1;
  else
    n1 = floor(n1);
  end
  if (~exist('n2', 'var'))
    n2 = 1;
  else
    n2 = floor(n2);
  end 
  if (~exist('n3', 'var'))
    n3 = 1;
  else
    n3 = floor(n3);
  end    
  if (~exist('n4', 'var'))
    n4 = 1;
  else
    n4 = floor(n4);
  end 
  if (~exist('n5', 'var'))
    n5 = 1;
  else
    n5 = floor(n5);
  end    
  
  if (n1 <= 1 )
    if  ( n2 > 1 )
      tmp=n1;  n1 = n2;  n2=tmp;
    elseif ( n3 > 1 )
      tmp=n1;  n1 = n3;  n3=tmp;
    elseif ( n4 > 1 )
      tmp=n1;  n1 = n4;  n4=tmp;  
    elseif ( n5 > 1 )
      tmp=n1;  n1 = n5;  n5=tmp; 
    else
      disp('**** Warning: inputs n1:n5 to conv_diff.m are all <=1');
      disp('**** replace n1 by 10');
      n1 = 10;
    end
  end
  
  if (~exist('potent', 'var'))
    potent=0; 
  end   
  if (~exist('alp1', 'var'))
    alp1 =0; 
  end
  if (~exist('alp2', 'var'))
    alp2 =0; 
  end
  if (~exist('alp3', 'var'))
    alp3 =0; 
  end
  if (~exist('alp4', 'var'))
    alp4 =0; 
  end
  if (~exist('alp5', 'var'))
    alp5 =0; 
  end  


  %
  % construct the matrix
  %
  d=2*ones(n1,1); dl=(-1+alp1)*ones(n1,1); du=(-1-alp1)*ones(n1,1);
  A = spdiags([dl, d, du], -1:1, n1, n1);

  if (n2 < 1) 
    n2 = 1;
  end
  if (n2 > 1) 
    d=2*ones(n2,1); dl=(-1+alp2)*ones(n2,1); du=(-1-alp2)*ones(n2,1);
    T = spdiags([dl, d, du], -1:1, n2, n2) ;
    A = kron(speye(n2,n2),A) + kron(T,speye(n1,n1)); 
  end
  
  if (n3 < 1),
    n3 = 1;
  end
  if (n3 > 1) 
    d=2*ones(n3,1); dl=(-1+alp3)*ones(n3,1); du=(-1-alp3)*ones(n3,1);
    T = spdiags([dl, d, du], -1:1, n3, n3);    
    A = kron(speye(n3,n3),A) + kron(T,speye(n1*n2)); 
  end
  
  if (n4 < 1),
    n4 = 1;
  end
  if (n4 > 1) 
    d=2*ones(n4,1); dl=(-1+alp4)*ones(n4,1); du=(-1-alp4)*ones(n4,1);
    T = spdiags([dl, d, du], -1:1, n4, n4);    
    A = kron(speye(n4,n4),A) + kron(T,speye(n1*n2*n3)); 
  end 
  
  if (n5 < 1),
    n5 = 1;
  end
  if (n5 > 1) 
    d=2*ones(n5,1); dl=(-1+alp5)*ones(n5,1); du=(-1-alp5)*ones(n5,1);
    T = spdiags([dl, d, du], -1:1, n5, n5);    
    A = kron(speye(n5,n5),A) + kron(T,speye(n1*n2*n3*n4)); 
  end 
   
  if (nargin >=6  &  abs(potent) > 0 ) 
    if (max(size(potent)) == 1),  %potent is a scalar shift
      A = A - potent * speye(size(A));
    else  %potent is a vector that stores the value of potential
      if (size(potent,1) ~= size(A,1)),
	error('dimension miss match between the matrix and the shift of potential')
      else
	%for (k= 1: size(A,1)) 
	%  A(k,k)=A(k,k)-potent(k);
	%end
	%--the following appears to be slow for large dim
	A = A - spdiags(potent, 0, size(A,1), size(A,1));
      end
    end
  end
  
  %
  % print out some info 
  %
  if (alp1^2+ alp2^2+ alp3^2+ alp4^2+ alp5^2== 0),
    if (abs(potent)==0),
      modeltype = 'Laplacian';
    else
      modeltype = 'Hamiltonian';
    end
  else
    %modeltype = 'Convection Diffusion';
    modeltype = 'Conv-Diff';
  end
  
  nd=sum([n1,n2,n3,n4,n5]>1);
  fprintf(' Model %2i-d  %s, dim=%i.  ', nd, modeltype, n1*n2*n3*n4*n5);
  if (nd > 1),
    fprintf(' Grid = ( %i', n1)
    if (n2 > 1),
      fprintf(' x %i', n2)
    end
    if (n3 > 1),
      fprintf(' x %i', n3)
    end   
    if (n4 > 1),
      fprintf(' x %i', n4)
    end
    if (n5 > 1),
      fprintf(' x %i', n5)
    end      
    fprintf(' )\n')
  else
    fprintf('\n')
  end
  
