function [v] = dgks(X, v0, orthtest);
% 
% Usage:  v = dgks( X, v0 ).
%
% Apply DGKS correction (two-step MGS) to ortho-normalize v0 against X,
% i.e.   v = (I - XX^T)*v0;  or  v = (I-XX^T)*(I-XX^T)*v0; 
%  and   v = v /norm(v);  
%  
% It is assumed that X is already ortho-normal, this is very important 
% for the projection  P = I - X*X^T  to be orthogonal.
%
% For debugging purpose, a 3rd variable can be provided to test if
% X is ortho-normal or not. This input variable can be of any value.
%
%--y.k. zhou
%  UMN, 2004
%
  
  if (nargin==3),
    if (orthtest==1),
      xorth=norm(X'*X - eye(size(X,2)));
      fprintf('   Orthgonality test:  ||X^t*X - I|| = %e\n', xorth)
    end
  end
  
  re_iterate = 1;
  
  while (re_iterate < 3)
%
% normalize v0 for the sake of numerical stability
% (this step can be important in case of badly scaled v0)
%
  nrmv0 = norm(v0);
  if (nrmv0 <= eps),
    fprintf('*** Warning: In dgks.m the input v0 is a zero vector\n')
    fprintf('*** Auto replace v0 by a random vector randn(size(v0))\n')
    v0 = randn(size(v0));
    v0 = v0/norm(v0);
  else
    v0 = v0 / nrmv0;
    nrmv0 = 1;
  end
  
%  
% the first step Gram-Schmidt: v=(I -XX')v0 
%  
  vtmp = X'*v0;
  v = v0 - X*vtmp;    
  nrmv = norm(v);  
  v = v/nrmv;  % always better to normalize (though it can be done later)
  % orth_test1 = norm(v'*X)

%
% the second step Gram-SchmidtS:  v =(I -XX')*v.
% a condition to check if a 2nd correction is necessary is used.
%
% note  tol = ctan(theta), 
% where theta is the angle between  X*vtmp and v
%
  tol = 1.41422;  % the angle is required to be no smaller than pi/4
                  % otherwise, a 2nd step orthogonalization is necessary

  nrmvtmp = norm(vtmp);		  
  if ( nrmvtmp*tol <= nrmv0 ),
    break;
  else
    % need 2nd step projection
    vtmp = X'*v;
    v = v - X*vtmp;   
    nrmv = norm(v); % use this value for the re-orth condition check
    v = v / nrmv;   
    %orth_test2 = norm(v'*X)
    
    if ( norm(vtmp) <= nrmv ), 
      break;
    else
      % need 3rd step projection
      fprintf(' *** Warning: two steps Gram-Schmidt is not enough! \n')
      vtmp = X'*v;
      v = v - X*vtmp;   
      nrmv = norm(v);  
      v = v / nrmv;
      orth_test3 = norm(v'*X);
      fprintf(' *** After 3rd step Gram-Schmidt, orth_test3=%e\n', orth_test3)
      if ( norm(vtmp) <= nrmv ),
	break;
      else 
	% the following normally doesn't happen, 3-step CGS should be enough
	fprintf('**** Error: three steps GS is not enough => \n')
	fprintf('     v0 is in the subspace it is orthogonalizing to. \n') 
	fprintf('**** Replace v0 by a random vector randn(size(v0))\n')
	v0 = randn(size(v0));  
	re_iterate = re_iterate +1;
      end
    end
    
  end
  if (re_iterate > 2),
    fprintf('\n*##* Error: in calling function [v] = dgks(X, v0)\n')
    fprintf('*##* 6 CGS steps failed, can not re-orth v0 against X\n\n');
  end
  end %while