function [eval, eigV, nconv, history] = gdav(varargin);
%
% gdav includes several Davidson-type methods for computing 
% 1. eigenpairs for symmetric/Hermitian matrices, or 
% 2. partial Schur decomposition for non symmetric/Hermitian matrices.
%
% Usage:
%      [eval, eigV, nconv, history] = gdav(varargin)
%
% where   gdav(varargin)  is one of the following:
%--------------------------------------------------------------------    
% 1.   gdav(A);
% 2.   gdav(A, nwant);
% 3.   gdav(A, nwant, mode);
% 4.   gdav(A, nwant, mode, opts);
%--------------------------------------------------------------------
% 
% A is the matrix for the eigenproblem, (dim=size(A,1)), 
%   (it is possible to input only a function for mv products).
%
% nwant is the number of wanted eigen pairs,
%   by default: nwant=5.
%
% mode is the wanted eigen-mode, it includes:
%   'SA' -- smallest algebraic
%   'LA' -- largest algebraic
%   'SR' -- smallest real part (same as 'SA' for symmetric/Hermitian problem)
%   'LR' -- largest real part (same as 'LA' for symmetric/Hermitian problem)
%   'SM' -- smallest magnitude
%   'LM' -- largest magnitude
%   'SI' -- smallest imaginary part
%   'LI' -- largest imaginary part
%   by default: mode='SR'. 
%  
% opts is a structure containing additional parameters, 
% the accepted field names are:
%        tol -- convergence tolerance for the residual norm of each eigenpair;
%               by default:  tol =1e-10.
%    Matsymm -- signal if the matrix is symmetric/Hermitian.
%               by default:  Matsymm=0.   
%               if Matsymm is not set and the input matrix is numeric, then
%               Matsymm = isequal(A,A').  
%       vmax -- maximum subspace dimension;
%               by default:  vmax= max(2*nwant, 20).
%      itmax -- maximum iteration number;
%               by default:  itmax= max(dim, 400).
%      nkeep -- number of vectors to keep during restart,
%               by default:  nkeep= nwant.
%         v0 -- the initial vector;
%               by default: v0 = rand(dim,1).
%      displ -- information display level;  by default: displ=1.
%               (<=0 --no output; 1--some output; >=2 --more output) 
%      alpha -- a numerical scalar as in equation (A- mu I + alpha x y') t =-r
%               (by default, alpha=2 and is auto adjusted if necessary)
%     nomore -- if set to 1 or logical(1), then only nwant eigenpairs are
%               computed. (by default, nomore==logical(0))
%
%    les_tol -- linear (correction) equation solve tolerance,
%               by default: les_tol=1e-3.
%  les_itmax -- linear (correction) equation solve maximum iteration number;
%               by default: les_itmax=7.
% les_solver -- solver used for the linear (correction) equations,
%               the default solver used is 31 (equivalently: 'jd-bicgs').
%               (the numbers chosen do not have specific meaning):
%      1 ('infl-gmres')   -- inflated RQI using  GMRES
%      2 ('cons-gmres')   -- constrained Newton using  GMRES
%      3 ('rand-gmres')   -- randomized RQI using  GMRES  
%      4 ('infl-bicgs')   -- inflated RQI using  BICGSTAB
%      5 ('cons-bicgs')   -- constrained Newton using  BICGSTAB
%      6 ('rand-bicgs')   -- randomized RQI using  BICGSTAB
%      7 ('infl-minres')  -- inflated RQI using  MINRES
%      8 ('consm-gmres')  -- constrained RQI using  GMRES
%      9 ('consm-bicgs')  -- constrained RQI using  BICGSTAB
%     10 ('rpx-gmres')    -- (r+x)   variant using GMRES
%     11 ('rpxa-gmres')   -- (r+x)*A variant using GMRES  
%     12 ('rpx-bicgs')    -- (r+x)   variant using BICGSTAB
%     13 ('rpxa-bicgs')   -- (r+x)*A variant using BICGSTAB
%     14 ('sym-gmres')    -- symmetric update using GMRES
%     15 ('sym-bicgs')    -- symmetric update using BICGSTAB
%
%     21 ('rqi-gmres')    -- approximate RQI using GMRES,
%     22 ('rqi-gmres-fp') -- approximate RQI using GMRES, full projection;
%     23 ('rqi-bicgs')    -- approximate RQI using BICGSTAB,
%     24 ('rqi-bicgs-fp') -- approximate RQI using BICGSTAB, full projection;
%     25 ('rqi-minres')   -- approximate RQI using MINES,
%     26 ('rqi-minres-fp')-- approximate RQI using MINES, full projection.
%
%     31 ('jd-bicgs')     -- JD  using  BICGSTAB,
%     32 ('jd-bicgs-fp')  -- JD  using  BICGSTAB, full projection. 
%     41 ('jd-gmres')     -- JD  using  GMRES,
%     42 ('jd-gmres-fp')  -- JD  using  GMRES, full projection.
%
%     51 ('jdm-bicgs')    -- JDm using  BICGSTAB,
%     52 ('jdm-bicgs-fp') -- JDm using  BICGSTAB, full projection.
%     61 ('jdm-gmres')    -- JDm using  GMRES,
%     62 ('jdm-gmres-fp') -- JDm using  GMRES, full projection.
%     71 ('jd-minres')    -- JD  using  MINRES,
%     72 ('jd-minres-fp') -- JD  using  MINRES, full projection.
%
%     note: for JD and JDm, the "full projection" means that the
%           projector is  (I - [V,u][V,u]') instead of (I - uu').
%
%--------------------------------------------------------------------------
%
%   NOTE that for different problems, the above defaults for les_itmax,
%   les_solver and les_tol need to be adjusted for better efficiency.
%
%  
%========== Output variables:
%
%    eval (optional): 
%         1. symmetric case:  eval returns a vector containing
%            the converged eigenvalues. 
%         2. non-symmetric case: eval returns the upper triangular
%            matrix of the converged partial Schur decomposition.
%
%    eigV (optional): [since it is always computed, not specifying this
%                      output does not save cputime]
%         1. symmtric case: converged eigenvectors.
%         2. non symmetric case: converged Schur vectors 
%           
%   nconv:  number of converged eigenvalues (optional).
%
% history:  log information (optional)
%           history(:,1) -- iteration number
%           history(:,2) -- cumulative number of matrix-vector products 
%                           at each iteration
%           history(:,3) -- residual norm at each iteration
%           history(:,4) -- current approximation of the wanted eigenvalues
%
%
%--------------------------------------------------------------------------
% As an example:
%
%    A = delsq(numgrid('D',90));   A=A-1.6e-2*speye(size(A));
%    opts = struct('les_solver', 31, 'les_itmax', 8, 'displ', 1);
%    [eval, eigV] = gdav(A, 10, 'SR', opts);  
%
% will compute 10 smallest real-part eigenpairs of A, using the specified
% values in opts and the defaults for the other unspecified parameters.
%

%  
%---y.k. zhou  (zhou@msi.umn.edu)
%   Nov, 2004, UMN
%

  global Matsymm            % Matsymm==1 if the matrix is symmetric.
  global MVprod             % count the matrix-vector products
  %global nprecsol          % count the preconditioned solves
  global alpha  is_alpha    % alpha as in (A - mu I + alpha x y') t = -r
  is_alpha = logical(0);    % will be logical(1) if alpha is provided
  
  %
  % Process inputs and do error-checking
  %
  % No input arguments, return help.
  if nargin == 0, help gdav, return, end
  if isnumeric(varargin{1})
    global A_operator
    A_operator = varargin{1}; 
    [dim]=size(A_operator,1);
    if (dim ~= size(A_operator,2)),
      error('The input numeric matrix must be a square matrix')
    end
    Anumeric = 1;
  else
    A_operator = fcnchk(varargin{1});
    Anumeric = 0;
    dim = varargin{2};
    if (~isnumeric(dim) | ~isreal(dim) | round(dim)~=dim | dim <1)
      error('A valid matrix dimension is required for the input string function')
    end
  end 

  if (Anumeric),
    Matsymm = isequal(A_operator,A_operator');
  else
    Matsymm = 0;  
  end

  if (dim <=200  & Anumeric), 
  %if (dim <=50 & Anumeric), 
    if (Matsymm==1),
      warning('small dimension problem, use eig() instead')
      [eigV, eval]=eig(full(A_operator));
      eval = diag(eval);
    else
      warning('small dimension problem, use schur() instead')
      [eigV, eval]=schur(full(A_operator));
    end
    if (nargout >2), nconv = dim; end
    if (nargout >3), history=[]; end
    return
  end

  %
  % Set default values or apply existing input options:
  %
  if (nargin < 3-Anumeric),
    nwant = min(dim,5);
  else
    nwant = varargin{3-Anumeric};
    if (~isnumeric(nwant) | nwant < 1 | nwant > dim),
      warning('invalid # of wanted eigenpairs input to gdav.m. use default value')
      nwant = min(dim,5);
    end
  end

  if (nargin < 4-Anumeric),
    mode = 'SR';
  else
    mode = upper(varargin{4-Anumeric});  mode=mode(1:2);
    if (Matsymm== 1),
      if (mode~='LM'&mode~='SM'&mode~='LA'&mode~='SA'&mode~='SR'&mode~='LR')
        warning('invalid wanted eigen-mode input to gdav.m. use default value')
	mode = 'SR';
      end
    else
      if (mode~='LM'&mode~='SM'&mode~='SR'&mode~='LR'&mode~='SI'&mode~='LI')
        warning('invalid wanted eigen-mode input to gdav.m. use default value')
	mode = 'SR';
      end      
    end
  end

  %
  % list all the rest default values, they will be overwritten 
  % if provided through structure opts
  %
  tol    = 1e-10;
  vmax   = max(2*nwant, 20);
  itmax  = max(dim, 400);
  nkeep  = nwant;
  v0     = rand(dim,1);
  displ  = 1;
  n_more = 4;       %by default, check 4 more eigenpairs (if necessary)
  les_tol= 1e-3;    %linear-equation-solve tolerance
  les_itmax  = 7;   %linear-equation-solve maximum iteration number
  les_solver = 'jd-bicgs';  %default solver: JD-bicgstab 
                            %(by no means is this default the best choice, it 
                            %is chosen only to give les_solver a default value)
			      
  if (nargin >= (5-Anumeric))
    opts = varargin{5-Anumeric};
    if ~isa(opts,'struct')
        error('Options argument must be a structure.')
    end
    
    if (isfield(opts,'Matsymm'))  
      Matsymm = opts.Matsymm;
      if  (Matsymm ~= 1)
	Matsymm =0;    %any opts.Matsymm ~=1 is treated as non-symmetric
      end
    end
    
    if isfield(opts,'tol')
      tol = opts.tol;
      if ~isequal(size(tol),[1,1]) | ~isreal(tol) | (tol<=0)
	warning('invalid tol input to gdav.m. use default value')
	tol = 1e-8;
      elseif  (tol < 1e-14)
	%the Davidson-type methods are not for eps accuracy computation.
	warning('tol input to gdav.m is too small. replace by 1e-13')	
	tol = 1e-13;
      end
    end
    
    if isfield(opts,'vmax')
      vmax = opts.vmax;
      if ~isequal(size(vmax),[1,1]) | ~isreal(vmax) | (vmax<1) | (vmax>dim)
	warning('invalid vmax input to gdav.m. use default value')
	vmax =  max(2*nwant, 20);
      end
    end
    
    if isfield(opts,'nkeep')
      nkeep = opts.nkeep;
      if ~isequal(size(nkeep),[1,1]) | ~isreal(nkeep) | (nkeep<1) | (nkeep>dim)
	warning('invalid nkeep input to gdav.m. use default value')
	nkeep =  nwant;
      end
    end
    
    if isfield(opts,'itmax')
      itmax = opts.itmax;
      if ~isequal(size(itmax),[1,1]) | ~isreal(itmax) | (itmax< 10)
	warning('invalid itmax input to gdav.m. use default value >=100')
	itmax = max(dim, 100);
      end
    end  
    
    if isfield(opts,'v0')
      v0 = opts.v0;
      if (size(v0,1)~=dim | ~isreal(v0) | norm(v0,1) < 2.2204e-16),
	warning('invalid v0 input to gdav.m. use default value')
	v0 = rand(dim, 1);
      end
    end   
    
    if isfield(opts,'displ')
      displ = opts.displ;
      if (~isequal(size(displ),[1,1])) | (~isreal(displ))
	warning('invalid displ input to gdav.m. use default value 1')
	displ =1;
      end
    end

    if isfield(opts,'les_itmax')
      les_itmax = opts.les_itmax;
      if (~isequal(size(les_itmax),[1,1]) | ~isreal(les_itmax) | les_itmax<0)
	warning('invalid les_itmax input to gdav.m. use default value 15')
	les_itmax =7;
      end
    end   
  
    if isfield(opts,'les_tol')
      les_tol = opts.les_tol;
      if (~isequal(size(les_tol),[1,1]) | ~isreal(les_tol) | les_tol<0)
	warning('invalid les_tol input to gdav.m. use default value 1e-3')
	les_tol =1e-3;
      end
    end  

    if isfield(opts,'nomore')
      nomore = opts.nomore;  %control if to compute more than nwant eig pairs
      if (nomore ~= 1 & nomore ~=logical(1)),
	n_more = 4;           %use the default value
      else
	n_more = 0;
      end
    end    
    
    if isfield(opts,'alpha')
      alpha = opts.alpha;   
      is_alpha = logical(1);
      if (~isnumeric(alpha)),
	warning('invalid alpha input to gdav.m. use default value alpha=2')
	is_alpha = logical(0);
      end
    end    
     
    if isfield(opts,'les_solver')
      les_solver = opts.les_solver;
      switch les_solver
       case {1, 'infl-gmres'},    les_solver= 'infl-gmres';
       case {2, 'cons-gmres'},    les_solver= 'cons-gmres';
       case {3, 'rand-gmres'},    les_solver= 'rand-gmres';	
       case {4, 'infl-bicgs'},    les_solver= 'infl-bicgs';
       case {5, 'cons-bicgs'},    les_solver= 'cons-bicgs';
       case {6, 'rand-bicgs'},    les_solver= 'rand-bicgs';
       case {7, 'infl-minres'},   les_solver= 'infl-minres';
       case {8, 'consm-gmres'},   les_solver= 'consm-gmres';	
       case {9, 'consm-bicgs'},   les_solver= 'consm-bicgs';	
       case {10, 'rpx-gmres'},    les_solver= 'rpx-gmres';	
       case {11, 'rpxa-gmres'},   les_solver= 'rpxa-gmres';
       case {12, 'rpx-bicgs'},    les_solver= 'rpx-bicgs';		
       case {13, 'rpxa-bicgs'},   les_solver= 'rpxa-bicgs';
       case {14, 'sym-gmres'},    les_solver= 'sym-gmres';	
       case {15, 'sym-bicgs'},    les_solver= 'sym-bicgs';		
	
       case {31, 'jd-bicgs'},     les_solver= 'jd-bicgs';  
       case {32, 'jd-bicgs-fp'},  les_solver= 'jd-bicgs-fp';
	
       case {71, 'jd-minres'},    les_solver= 'jd-minres';     
       case {72, 'jd-minres-fp'}, les_solver= 'jd-minres-fp';
            
       case {41, 'jd-gmres'},     les_solver= 'jd-gmres';
       case {42, 'jd-gmres-fp'},  les_solver= 'jd-gmres-fp';  
      
       case {51, 'jdm-bicgs'},    les_solver= 'jdm-bicgs';   
       case {52, 'jdm-bicgs-fp'}, les_solver= 'jdm-bicgs-fp'; 
	 
       case {61, 'jdm-gmres'},    les_solver= 'jdm-gmres';   
       case {62, 'jdm-gmres-fp'}, les_solver= 'jdm-gmres-fp';

       case {21, 'rqi-gmres'},    les_solver= 'rqi-gmres';   
       case {22, 'rqi-gmres-fp'}, les_solver= 'rqi-gmres-fp'; 
       case {23, 'rqi-bicgs'},    les_solver= 'rqi-bicgs';    
       case {24, 'rqi-bicgs-fp'}, les_solver= 'rqi-bicgs-fp';
       case {25, 'rqi-minres'},   les_solver= 'rqi-minres'; 	
       case {26, 'rqi-minres-fp'},les_solver= 'rqi-minres-fp'; 
	
       otherwise
	 warning('invalid les_solver input to gdav.m. use default')
	 les_solver = 'jd-bicgs';
      end
    end  
  end
  
  % Finally finished the boring checks on input variables.  
  
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  %
  % Now start the main algorithm:
  %
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
  %Preallocate memory:
   resnrm=zeros(nwant,1);
   V = zeros(dim, vmax); 
   W = zeros(dim, vmax); 
   Hn= zeros(vmax,vmax);  
   if (Matsymm~=1), R=zeros(nwant,nwant); end
   eval=zeros(nwant,1);  %over written by R in the non-symmetric case 

   
   if (nargout > 3),  
     longlog = 1;  % produce long logs in history(:,1:4)  
   else            
     longlog = 0;
   end 

   nconv = 0;      % init number of converged eigenvalues
   nconv1= 1;      % nconv1 should always be nconv+1
   MVprod = 0;     % initialize the MV-product count
   nit_total = 1;  % init the total iteration number count
   CHECKALL=1;     % CHECKALL~=1 can be a more robust choice

  
   [x, rt] = qr(v0, 0); 
   V(:,1) = x;   

   if (Anumeric),
     W(:,1)  = user_Hx( x );
   else
     W(:,1)  = user_Hx( x, A_operator );
   end
   rho =  x' * W(:,1);
   Hn(1,1) = rho;
   r = W(:,1) - Hn(1,1) * x;
   
   normr = norm(r);
   %beta = abs(rho)+normr;
   beta  = max(2.25e-14,abs(rho));
   tolr  = tol*beta;               %set initial tolerance
   if (normr < tol) 
     nconv  = nconv +1;
     nconv1 = nconv +1;
     eval(nconv)  = rho;
   end
   if (displ > 1),
     if (imag(rho) ~= 0),
       fprintf (' initial rho=(%e,%e),  rnorm = %e\n',  real(rho), ...
		imag(rho), normr);
     else
       fprintf (' initial rho=%e,  rnorm = %e\n', rho, normr);
     end
   end
   
   n = 1;  %n stores the subspace dimension (NOTE n is NOT the dimension dim)

   if (longlog == 1),
       history(1, 1) = 1;
       history(1, 2) = MVprod;
       history(1, 3) = normr;
       history(1, 4) = rho;
       nlog = 2;
   end      
   
   while (nit_total <= itmax) 
     n1 = n +1;  % n1 should always be n+1, 
		 % n1 is the subspace dimension after augmentation at this step
     nit_total = nit_total +1;

     if (Anumeric),
      switch (les_solver)
      case {'infl-bicgs', 'infl-gmres', 'infl-minres', 'rand-gmres', ...
	    'rand-bicgs','cons-gmres','cons-bicgs','consm-gmres', ...
	    'consm-bicgs','rpx-gmres','rpx-bicgs','rpxa-gmres', ...
	    'rpxa-bicgs', 'sym-gmres','sym-bicgs'},
         if (is_alpha),
	   t=modrqi(r,rho,x,les_itmax,les_tol, les_solver, alpha);
	 else
	   t=modrqi(r,rho,x,les_itmax,les_tol, les_solver);
	 end
	 
      case {31, 'jd-bicgs'}      % iterative solve by JD-bicgstab
         t=jd_solve(r,rho,x,les_itmax,les_tol,'bicgstab',V(:,1:nconv));
      case {32, 'jd-bicgs-fp'}   % iterative solve by JD-bicgstab-FP
         t=jd_solve(r,rho,x,les_itmax,les_tol,'bicgstabFP',V(:,1:nconv));
	 
      case {71, 'jd-minres'}     %iterative solve by JD-minres (symmetric case)
       if (Matsymm==1),
         t=jd_solve(r, rho, x, les_itmax, les_tol,'minres',V(:,1:nconv));
       else
	 error('MINRES works only for symmetric problems')
       end
      case {72, 'jd-minres-fp'}  %iterative solve by JD-minres-FP (symmetric case)
       if (Matsymm==1),
         t=jd_solve(r, rho, x, les_itmax, les_tol,'minresFP',V(:,1:nconv));
       else
	 error('MINRES works only for symmetric problems')
       end 
            
      case {41, 'jd-gmres'}      % iterative solve by JD-gmres
         t=jd_solve(r, rho, x, les_itmax, les_tol,'gmres', V(:,1:nconv));
      case {42, 'jd-gmres-fp'}   % iterative solve by JD-gmres-FP
         t=jd_solve(r, rho, x, les_itmax, les_tol,'gmresFP',V(:,1:nconv));
      
      case {51, 'jdm-bicgs'}     % iterative solve by JDm-bicgstab
         t=jdm_solve(r, rho, x,les_itmax,les_tol,'bicgstab',V(:,1:nconv));
      case {52, 'jdm-bicgs-fp'}  % iterative solve by JDm-bicgstab-FP
         t=jdm_solve(r, rho, x,les_itmax,les_tol,'bicgstabFP',V(:,1:nconv));
	 
      case {61, 'jdm-gmres'}     % iterative solve by JDm-gmres
         t=jdm_solve(r, rho, x, les_itmax, les_tol,'gmres',V(:,1:nconv));
      case {62, 'jdm-gmres-fp'}  % iterative solve by JDm-gmres-FP
         t=jdm_solve(r, rho, x, les_itmax, les_tol,'gmresFP',V(:,1:nconv));

      case {21, 'rqi-gmres'}     % iterative solve by RQI-gmres
         t=itsolve_rqi(r,rho,x,les_itmax,les_tol,'gmres');
      case {22, 'rqi-gmres-fp'}  % iterative solve by RQI-gmres-FP
         t=itsolve_rqi(r,rho,x,les_itmax,les_tol,'gmres', ...
			 V(:,1:nconv));	
      case {23, 'rqi-bicgs'}     % iterative solve by RQI-bicgstab
         t=itsolve_rqi(r,rho,x,les_itmax,les_tol,'bicgstab');
      case {24, 'rqi-bicgs-fp'}  % iterative solve by RQI-bicgstab-FP
         t=itsolve_rqi(r,rho,x,les_itmax,les_tol,'bicgstab', ...
			 V(:,1:nconv));
      case {25, 'rqi-minres'}    %iterative solve by RQI-minres (symmetric case)
        if (Matsymm==1),
	  t=itsolve_rqi(r,rho,x,les_itmax,les_tol,'minres');
	else
	  error('MINRES works only for symmetric problems')
	end 
      case {26, 'rqi-minres-fp'} %iterative solve by RQI-minres-FP (symmetric case)
        if (Matsymm==1),
	  t=itsolve_rqi(r,rho,x,les_itmax,les_tol,'minres', ...
			 V(:,1:nconv));
	else
	  error('MINRES works only for symmetric problems')
	end 
      
      otherwise
         error('Input les_solver type to  gdav.m  does not exist')
      end
     
     else % A_operator is a script for matrix-vector-product,
          % for this case the interfaces have not been tested extensively yet
      t= solveA(r,rho,x,les_itmax,les_tol, les_solver, V(:,1:nconv), A_operator);
     end
     
     %
     % make t orthogonal to all vectors in V (including the converged V(:,1:nconv))
     %
     V(:, n1) =  dgks( V(:, 1:n), t );
     
     %
     % do the new matrix-vector product.
     %
     if (Anumeric),
       W(:, n1) =  user_Hx( V(:, n1) );
     else
       W(:, n1) =  user_Hx( V(:, n1), A_operator );  
     end     

     
     %
     % compute only the active part (not including the deflated part)
     %
     Hn(nconv1:n1, n1) = V(:, nconv1:n1)'* W(:, n1); 
     if (Matsymm == 1), % the symmetric case 
       Hn(n1, nconv1:n) = Hn(nconv1:n, n1)';  
     else               % the non-symmetric case 
       Hn(n1, nconv1:n) = V(:,n1)'*W(:, nconv1:n);
     end

     if (Matsymm==1),
       [Eig_vec, Eig_val] = eig(Hn(nconv1:n1,nconv1:n1));  
       d_eig = diag(Eig_val);
       
       if (mode =='SR' | mode =='SA')
	 [d_eig_tmp, indx] = sort(real(d_eig));  
       elseif (mode =='LR'| mode =='LA')
	 [d_eig_tmp, indx] = sort(-real(d_eig));     
       elseif (mode =='SM'), 
	 [d_eig_tmp, indx] = sort(abs(d_eig));
       elseif (mode =='LM'), 
	 [d_eig_tmp, indx] = sort(-abs(d_eig));
       elseif (mode =='SI' |  mode =='LI' )
       if (Matsymm==1),
        error('**** symmetric/Hermitian eigenproblem, no LI or SI mode fits')
       else
        if  ( mode =='LI' )
          [d_eig_tmp, indx] = sort(-imag(d_eig));
        elseif  ( mode =='SI' )
          [d_eig_tmp, indx] = sort(imag(d_eig));
        end
       end
       else
	 error('the input wanted eigen-mode to gdav.m is invalid')
       end
       % reorder Ritz pairs
       d_eig = d_eig(indx);  
       Eig_vec=Eig_vec(:, indx); 
       
     else
       %
       % the non-symmetric case, need to use Schur decomposition
       % note that now (Eig_vec, Eig_val) contains Schur pairs
       %
       [Eig_vec, Eig_val] = schur(Hn(nconv1:n1,nconv1:n1));
       [Eig_vec, Eig_val, indx] = sortschur(Eig_vec, Eig_val, mode);
       d_eig = diag(Eig_val);
     end

     
     if ( n1 >= vmax ),  
       % need to restart, note that only n1 need to be updated
       n1 = max([nconv1,  nwant+5,  min(nkeep + nconv, vmax-5)]);
     end
     
     %
     % it is more efficient to do a Rayleigh-Ritz refinement at each step
     % (with this extra work, certain eigenvectors may converge faster)
     % Moreover, this refinement makes  Hn=V'AV  diagonal (for Matsymm==1)
     % or uppertriangular (for Matsymm~=1);
     % otherwise the latter deflation won't work properly.
     %
     % by placing the restart test above, one can just keep the necessary
     % columns of V and W.
     % (note that at this step, n still stores the old value)
     %
     icount = n1-nconv1+1;
     if (Matsymm==1),
       Hn(nconv1:n1,nconv1:n1) = diag(d_eig(1:icount));
     else
       %keep the reordered upper triangular part of Eig_val
       Hn(nconv1:n1,nconv1:n1)=Eig_val(1:icount, 1:icount);
     end
     %
     % for nonsymmetric problems, in order not to lose the V'V=I,
     % the Eig_vec has to contain the Schur vectors.
     %
     V(:,nconv1:n1)=V(:,nconv1:n+1)*Eig_vec(:,1:icount);
     W(:,nconv1:n1)=W(:,nconv1:n+1)*Eig_vec(:,1:icount);


     beta1 = max(abs(d_eig));
     %--deflation and restart may make beta1 too small, (the active subspace
     %--dim is small at the end), so use beta1 only when it is larger.
     if (beta1 > beta),
       beta = beta1; 
       tolr = tol*beta;
     end
     
     rho = d_eig(1);
     x = V(:,nconv1);             %% x = x/norm(x);
   
     r = W(:, nconv1) - rho*x;   
     if (Matsymm~=1 & nconv > 0), % handle the nonsymmetric case
        rtmp=V(:,1:nconv)'*r;     
        r = r - V(:,1:nconv)*rtmp;
     end
     normr = norm(r);
     if (displ > 1),
       if (imag(rho) ~= 0),
	 fprintf (' ndim=%i, rho=(%e,%e),  rnorm = %e\n', n, real(rho), ...
		  imag(rho), normr);
       else
	 fprintf (' ndim=%i, rho=%e,  rnorm = %e\n', n, rho, normr);
       end
     end
     swap = logical(0);

     if ( CHECKALL == 1 )
       icount = n1 - nconv1+1;   % this count may be too large and can 
                                 % lead to missed eigenvalues
       icount = min(icount, 4);  % opt to check at most 4 Ritz pairs
     else                                            
       icount=1;  % reduce to check convergence of only one Ritz pair 
                  % at each step (this is a safer choice)
     end
     
     if (longlog == 1),
       history(nlog, 1) = nit_total;
       history(nlog, 2) = MVprod;
       history(nlog, 3) = normr;
       history(nlog, 4) = rho;
       nlog = nlog+1;
     end
     
     for ii =  1 : icount

       if  ( normr < tolr )
         nconv = nconv +1;
         nconv1 = nconv +1;
	 if (displ > 0),
	   if (imag(rho)==0)
	     fprintf('#%3i converged in %5i steps, ritz(%3i)=%e, rnorm=%6.4e\n',...
		      nconv,  nit_total, nconv,  rho, normr)
	   else
	     fprintf('#%3i converged in %5i steps, ritz(%3i)=(%e, %e), rnorm=%6.4e\n',...
		    nconv,  nit_total, nconv,  real(rho), imag(rho),  normr)
	   end
	 end
         eval(nconv) = rho; 
         resnrm(nconv) = normr;

	 if (Matsymm==1),
	 %
         % need to sort converged eigenvalues. 
	 % (the following can be easily simplified to handle only
	 %  symmetric cases)
	 %
	 % if ('SA' or 'SR' or 'SI'), sort in non-decreasing order.
         %     i.e.,  move smaller  converged eigenvalue value to the beginning. 
	 % if ('LA' or 'LR' or 'LI'), sort in non-increasing order.
	 % 
	 if (nconv >1)
	   vtmp = V(:,nconv);
	 end
         for i = nconv -1: -1 : 1
               if (mode=='SR'|mode=='SA') & real(rho) >= real(eval(i)),
                 break
	       elseif (mode=='LR'|mode=='LA') & real(rho) <= real(eval(i)),
		 break
	       elseif (mode=='SM') & abs(rho) >= abs(eval(i)),
		 break
	       elseif (mode=='LM') & abs(rho) <= abs(eval(i)),
		 break
	       elseif (mode=='SI') & imag(rho) >= imag(eval(i)),
		 break
	       elseif (mode=='LI') & imag(rho) <= imag(eval(i)),
		 break 
               else
                 eval(i+1)=eval(i);
                 eval(i) = rho;
		 V(:,i+1)= V(:,i); V(:,i)=vtmp;
                 swap = logical(1);
		 if (displ > 1),
		   fprintf(' *** swap %3i  with %3i\n', i+1, i);
		 end
               end
         end
	 
	 %--normally, do not continue if swap happens after nwant+n_more 
         if ( nconv >= nwant+n_more  & swap ),   
              swap = logical(0);
         end
	 else % non-symmetric case
	   R(nconv,nconv)=rho;
	   if(nconv>1), R(1:nconv-1,nconv)=rtmp; end
	 end

         if (nconv >= nwant & (~swap) ),
	   if (displ > 0),
	     fprintf('  The converged eigenvalues and residual_norms are:\n')
	     for i = 1 : nconv
	       if (imag(eval(i)) == 0),
		 fprintf('  eigval(%3i) = %11.8e,   resnrm(%3i) = %8.5e \n', ...
			  i,  eval(i), i,  resnrm(i))
	       else
		 fprintf('  eigval(%3i)=(%9.6e, %9.6e),  resnrm(%3i)=%8.5e\n', ...
			i,  real(eval(i)), imag(eval(i)), i,  resnrm(i))
	       end
	     end
	   end
	   % the eigV array can be saved. but in matlab, one can not just
	   % return V(:, 1:nconv) in the output argument list.
	   % for other languages, the following step should be unnecessary.
	   eigV = V(:, 1:nconv);

	   if (Matsymm~=1),
	     [eigV, eval, indx]= sortschur(eigV, R, mode);
	   end
	   if (displ > 0)
	     fprintf(' ====== linear solver used is:  %s\n\n', les_solver)
	   end
           return
         else
           % update rho, r and x to correspond to the next wanted eigenpair.
	   % if not convergent, the (rho, x) will be the ritz-pair to continue
	   % the iteration. (this includes the restart case)
           rho = d_eig(ii+1);
           x = V(:, nconv1);
           r = W(:, nconv1)  - rho *x;
	   if (Matsymm~=1 & nconv > 0),
	     rtmp=V(:,1:nconv)'*r;     
	     r = r - V(:,1:nconv)*rtmp;
	   end
	   normr = norm(r);
	   
	   if (longlog == 1 & normr < tolr),
	     history(nlog, 1) = nit_total;
	     history(nlog, 2) = MVprod;
	     history(nlog, 3) = normr;
	     history(nlog, 4) = rho;
	     nlog = nlog+1;
	   end
         end
       else
	 break;  % exit whenever the first non-convergent Ritz pair is reached
       end
     end

     n = n1;
        
     if ( nit_total > itmax),
       fprintf('**** itmax=%i,  it_total=%i\n', itmax, nit_total)
       warning(' **** gdav.m:  Maximum iteration exceeded')
       break
     end

   end

%
%-------------------------------------------------------------------------------
% Known problem:  
% 1. for non-hermitian problems,  when the convergence tol is too large, 
%  some real eigenvalues may have imaginary part at the order of  tol.
%  This is one disadvantage not using real arithmetic when A is real.
%  In Davidson-type methods, applying real-double-shift for a complex eigenvalue 
%  is not as convenient as that is in the (truncated) QR algorihm.
% 2. for non-hermitian problems with conjugate eigenvalues, the les_solver
%  named rqi-xxx do not work as well as other methods for nwant >1. 
% 3. all the interfaces have not been tested (some has been tested only roughly)
%  for the cases when the A_operator is a script for matrix-vector-product. 
%  (but this code is tested rather extensively when the matrix is numeric.) 
%
%-------------------------------------------------------------------------------
%
% Notice: this code is originally targeted for symmetric eigenproblems.
% It was later modified to solve both symmetric and non-symmetric problems.
% The modification is not quite polished. The code could be made more
% succinct, and variable names like (eval, EigV) etc that are for symmetric
% problems could be changed to more meaningful names for non-symmetric problems.
%
%