function [ z ] = itsolve_rqi(r, lam, x, itmax, tol, lsolver, eigV, Amat);
%
% solve the eq.  (Amat - lam B) z = x  using 
% iterative solver bicgstab, gmres and minres, 
% with different tol and itmax.
%   
% then set  x= z/tau - x (note, we are not solving (Amat-lamB)x=r+x/tau)
% 
% this is essentially the approximate Rayleigh-Quotient-Iteration with
% an additional orth projection at the end
%
  
  n = size(r,1);
  switch lsolver 
   case {'bicgstab'},
    if (nargin < 8),
      [z]= tk_bicgstab(@Kjv, x, tol, itmax, lam);
    else
      [z]= tk_bicgstab(@KjvA, x, tol, itmax, lam, Amat);
    end
   case {'gmres'},
    if (nargin < 8),
      [z, flag]= gmres(@Kjv, x, n, tol, itmax, [], [], [], lam);
    else
      [z, flag]= gmres(@KjvA, x, n, tol, itmax, [], [], [], lam, Amat);
    end
   case ('minres'),    %works only if Amat is symmetric
    if (nargin < 8),
      [z, flag]= minres(@Kjv, x, tol, itmax, [], [], [], lam);
    else
      [z, flag]= minres(@KjvA, x, tol, itmax, [], [], [], lam, Amat);
    end
  end
  
  nrmz=norm(z);
  if (nrmz == 0),
    % this step is for robustness, one does not want to quit simply
    % because one step fails (when Matlab linear solver returns
    % a zero vector).
    fprintf('*** Warning: in itsolve_rqi.m, solver=%s, ||z||=%e\n', ...
	    lsolver, nrmz); 
    fprintf('*** This is caused by the Matlab builtin solver used \n')
    fprintf('*** returned a zero vector instead of a \n'),
    fprintf('*** non-zero vector at the exit iteration step\n')
    fprintf('=== Now: replace solution z by a random vector\n')
    z = randn(size(x));
  end    
  
  if (exist('eigV', 'var')),
    % ortho against [eigV, x], can likely be better than the above.
    % this has not been tested in my submitted paper 
    x = dgks(eigV, x);
    tau = [eigV, x]'*z;
    z = z - [eigV, x]*tau;
    %z = z/ norm(z);
  else
    % only orth against x
    tau = x'*z;
    z = z/tau - x; %this works generally very well !
  end
  


%----------------------------------------------------------

function [y]=Kjv(r, lam)
%
% compute  y = (Amat-lamI)*r
%
  y = user_Hx(r);
  y = y - lam*r;

  
  
function [y]=KjvA(r, lam, Amat)
%
% compute  y = (Amat-lamI)*r
%
  y = user_Hx(r, Amat);
  y = y - lam*r;


  
  