function [ z ] = jd_solve(r, lam, x, itmax, tol, lsolver, eigv, A);
%
% Usage:
%    [z] = jd_solve(r, lam, x, itmax, tol, lsolver, eigv, A)
%
% solve the eq.  (I-xx')(A - lam I)(I-xx')z = r  using 
% iterative solver bicgstab,  gmres and minres, 
% with different tol and itmax.
%
% the input "eigV" is optional, it is used only for full projection.
% the input matrix "A" is also optional, it is mainly used when
% A is a script for the matrix-vector-product
%
  
n = size(r,1);
if (nargin < 8),
  switch lsolver 
   case {'bicgstab'},
    [z]= tk_bicgstab(@Kjv, r, tol, itmax, lam, x);
   case {'bicgstabFP'},
    [z]= tk_bicgstab(@KjvFP, r, tol, itmax, lam, x, eigv);
    
   case {'gmres'},
    %[z]= tk_gmres(@Kjv, r, tol, itmax, lam, x);
    [z, flag]= gmres(@Kjv, r, n, tol, itmax, [], [], [], lam, x);
   case {'gmresFP'},
    [z, flag]= gmres(@KjvFP, r, n, tol, itmax, [], [], [], lam, x, eigv);
  
   case ('minres'),    %works only if A is symmetric
    [z, flag]= minres(@Kjv, r, tol, itmax, [], [], [], lam, x);
   case ('minresFP'),  %works only if A is symmetric
    [z, flag]= minres(@KjvFP, r, tol, itmax, [], [], [], lam, x, eigv);
  end
else
  switch lsolver 
   case {'bicgstab'},
    [z]= tk_bicgstab(@KjvA, r, tol, itmax, lam, x, A);
   case {'bicgstabFP'},
    [z]= tk_bicgstab(@KjvFPA, r, tol, itmax, lam, x, eigv, A);
    
   case {'gmres'},
    %[z]= tk_gmres(@KjvA, r, tol, itmax, lam, x, A);
    [z, flag]= gmres(@KjvA, r, n, tol, itmax, [], [], [], lam, x, A);
   case {'gmresFP'},
    [z, flag]= gmres(@KjvFPA, r, n, tol, itmax,[], [], [], lam, x, eigv, A);
  
   case ('minres'),    %works only if A is symmetric
    [z, flag]= minres(@KjvA, r, tol, itmax, [], [], [], lam, x, A);
   case ('minresFP'),  %works only if A is symmetric
    [z, flag]= minres(@KjvFPA, r, tol, itmax, [], [], [], lam, x, eigv, A);
  end    
end

  nrmz=norm(z);
  if (nrmz == 0),
    %error('****in jd_solve, solver=%s: nrmz = %e\n', lsolver, nrmz);
    fprintf('*** Warning: in jd_solve.m, solver=%s, ||z||=%e\n', ...
	    lsolver, nrmz); 
    fprintf('*** This is caused by the solver used \n')
    fprintf('*** returned a zero vector instead of a \n'),
    fprintf('*** non-zero vector at the exit iteration step\n')
    fprintf('=== Now: replace solution z by a random vector\n')
    z = randn(size(x));
  end    

  tau = x'*z;
  z = z - tau*x;
 
  
%===========================================
function [y]=Kjv(r, lam, x)
%
% compute  y = (I-xx')(A-lamI)(I-xx')*r
%

  y = r - x*(x'*r);
  w = user_Hx (y);  %get w = A*y;
  y = w - lam*y;
  y = y - x*(x'*y);
  
  
%===========================================
function [y]=KjvFP(r, lam, x, eigv)
%
% compute  y = (I-QQ')(A-lamI)(I-QQ')*r
% where     Q = [eigv, x]
% FP in the function name means full projection, i.e.,  r is also projected
% to be orthogonal to the converged eigenvectors eigv.
% (this full projection actually does not gain much in practice)
%
  
  x  = dgks(eigv, x);
  vx = [eigv, x];   
  y = r - vx*(vx'*r);
  w = user_Hx (y);  %get w = A*y;
  y = w - lam*y;
  y = y - vx*(vx'*y);    
 
  
  
%-------------------------------------------------------------------
% the following are mainly useful for A not in numeric form
%-------------------------------------------------------------------
  
%===========================================
function [y]=KjvA(r, lam, x, A)
%
% compute  y = (I-xx')(A-lamI)(I-xx')*r
%
  y = r - x*(x'*r);
  w = user_Hx (y, A);  %get w = A*y;
  y = w - lam*y;
  y = y - x*(x'*y);
  
  
%===========================================
function [y]=KjvFPA(r, lam, x, eigv, A)
%
% compute  y = (I-QQ')(A-lamI)(I-QQ')*r
% where     Q = [eigv, x]
% FP in the function name means full projection, i.e.,  r is also projected
% to be orthogonal to the converged eigenvectors eigv.
% (this full projection actually does not gain much in practice)
%
  x  = dgks(eigv, x);
  vx = [eigv, x];   
  y = r - vx*(vx'*r);
  w = user_Hx (y, A);  %get w = A*y;
  y = w - lam*y;
  y = y - vx*(vx'*y);    
    