function [ z ] = jdm_solve(r, lam, x, itmax, tol, lsolver, eigv, A);
%
% Usage:
%    [z] = jdm_solve(r, lam, x, itmax, tol, lsolver, eigv, A)
%
% solve the eq.  (I-xx')(A - lam I) z = r  using 
% iterative solver bicgstab, gmres 
% with different tol and itmax.
%
% the input "eigV" is optional, it is used only for full projection.
% the input matrix "A" is also optional, it is mainly used when
% A is a script for the matrix-vector-product
%
  
n = size(r,1);
if (nargin < 8),
  switch lsolver 
   case {'bicgstab'},
    [z]= tk_bicgstab(@Kmodv, r, tol, itmax, lam, x);
   case {'bicgstabFP'},
    [z]= tk_bicgstab(@KmodvFP, r, tol, itmax, lam, x, eigv);
    
   case {'gmres'},
    %[z]= tk_gmres(@Kmodv, r, tol, itmax, lam, x);
    [z, flag]= gmres(@Kmodv, r, n, tol, itmax, [], [], [], lam, x);
   case {'gmresFP'},
    [z, flag]= gmres(@KmodvFP, r, n, tol, itmax, [], [], [], lam, x, eigv);
  
   case ('minres'),   %usually minres does not work in this case
    [z, flag]= minres(@Kmodv, r, tol, itmax, [], [], [], lam, x);
   case ('minresFP'), %usually minres does not work in this case
    [z, flag]= minres(@KmodvFP, r, tol, itmax, [], [], [], lam, x, eigv);
  end
else
  switch lsolver 
   case {'bicgstab'},
    [z]= tk_bicgstab(@KmodvA, r, tol, itmax, lam, x, A);
   case {'bicgstabFP'},
    [z]= tk_bicgstab(@KmodvFPA, r, tol, itmax, lam, x, eigv, A);
    
   case {'gmres'},
    %[z]= tk_gmres(@KmodvA, r, tol, itmax, lam, x, A);
    [z, flag]= gmres(@KmodvA, r, n, tol, itmax, [], [], [], lam, x, A);
   case {'gmresFP'},
    [z, flag]= gmres(@KmodvFPA, r, n, tol, itmax, [], [], [], lam, x, eigv, A);
  
   case ('minres'),   %usually minres does not work in this case
    [z, flag]= minres(@KmodvA, r, tol, itmax, [], [], [], lam, x);
   case ('minresFP'), %usually minres does not work in this case
    [z, flag]= minres(@KmodvFPA, r, tol, itmax, [], [], [], lam, x, eigv, A);  
  end
end

  nrmz=norm(z);
  if (nrmz == 0),
    %error('****in jdm_solve, solver=%s: nrmz = %e\n', lsolver, nrmz);
    fprintf('*** Warning: in jdm_itsolve.m, solver=%s, ||z||=%e\n', ...
	    lsolver, nrmz); 
    fprintf('*** This is caused by the Matlab builtin solver used \n')
    fprintf('*** returned a zero vector instead of a \n'),
    fprintf('*** non-zero vector at the exit iteration step\n')
    fprintf('=== Now: replace solution z by a random vector\n')
    z = randn(size(x));
  end    

  %note that the following orthogonalization is not performed
  %% tau = x'*z;
  %% z = z - tau*x;
 
  
%===========================================
function [y]=Kmodv(r, lam, x)
%
% compute  y = (I-xx')(A-lamI)*r
%
  %note the right projection is not performed
  %% y = r - x*(x'*r);
  y = user_Hx(r);  %get y = A*r;
  y = y - lam*r;
  y = y - x*(x'*y);
  
  
%===========================================
function [y]=KmodvFP(r, lam, x, eigv)
%
% compute  y = (I-QQ')(A-lamI)*r
% where     Q = [eigv, x]
% FP in the function name means full projection, i.e.,  r is also projected
% to be orthogonal to the converged eigenvectors eigv.
% (this full projection actually does not gain much in practice)
%
  
  x  = dgks(eigv, x);
  vx = [eigv, x];   
  % (no projection on the right side) 
  %% y = r - vx*(vx'*r);
  y = user_Hx(r);  %get y = A*r;
  y = y - lam*r;
  y = y - vx*(vx'*y);    
   
  
  
%===========================================
function [y]=KmodvA(r, lam, x, A)
%
% compute  y = (I-xx')(A-lamI)*r
%
  %note the right projection is not performed
  %% y = r - x*(x'*r);
  y = user_Hx(r, A);  %get y = A*r;
  y = y - lam*r;
  y = y - x*(x'*y);
  
  
%===========================================
function [y]=KmodvFPA(r, lam, x, eigv, A)
%
% compute  y = (I-QQ')(A-lamI)*r
% where     Q = [eigv, x]
% FP in the function name means full projection, i.e.,  r is also projected
% to be orthogonal to the converged eigenvectors eigv.
% (this full projection actually does not gain much in practice)
%
  x  = dgks(eigv, x);
  vx = [eigv, x];   
  % (no projection on the right side) 
  %% y = r - vx*(vx'*r);
  y = user_Hx(r, A);  %get y = A*r;
  y = y - lam*r;
  y = y - vx*(vx'*y);   