function [ z ] = modrqi(r, lam, x, itmax, tol, lsolver, alpha_in, A);
%
% Usage:
%    [z] =  modrqi(r, lam, x, itmax, tol, lsolver, alpha, A)
%
% solve the very general RQI-type correction equation:
%
%       (A - lam I + alpha x y')z = -r  
%
% using  iterative solver bicgstab, gmres, and minres (if possible), 
%
%
% for this script, choice of the vector y is controlled by the input
% lsolver. i.e.,  the beginning characters of lsolver control the 
% following:
%
%    "infl-*"  ==>   (A - lam I + alpha x x') z =-r
%
%    "rand-*"  ==>   y=rand(n,1);  (A - lam I + alpha x y')z =-r
%
%    "cons-*"  ==>   (A - lam I - 2 x (A*x)')z =-r 
%
%    "consm-*" ==>   (A - lam I + alpha x (A*x)')z =-r   (alpha~=-2)
%
%    "rpx-*"   ==>   (A - lam I + alpha x (r+ beta x)' )z =-r 
%    "rpxa-*"  ==>   (A - lam I + alpha x (r+ beta x)'A)z =-r   
%
%    "sym-*"   ==>   (A - lam I - r*x'-x*r'- alpha x*x')z =-r 
%                    (symmetric update, poor performance, note that
%                     this one does not observe the [alpha x y'] form)   
%
% (the parameter alpha is rather free, it can be any value that makes
%  the matrix non-singular. note that alpha is optional)
%  
% the input matrix "A" is optional, it is mainly used when A is a 
% script for the matrix-vector-product
%

%
% y.k. zhou  (zhou@msi.umn.edu)
% may-2005
%

global A_operator  Matsymm  
%only used for the x'(A+A') term in Kcons when A is not symmetric
		    
global  alpha_loc  %alpha_loc will be overwritten when necessary

if (exist('alpha_in', 'var')),
  alpha_loc = alpha_in;
  if (alpha_loc==0), alpha_loc=10; end 
else
  alpha_loc=10;  %alpha defaults to 10 (can be one of many other values) 
end

%it is found that alpha_loc be at the same order of lam is good when 
%lam is large (say,  |lam|>1e+3)
if (abs(lam) > alpha_loc),
  alpha_loc = - lam;        %note that the minus sign does not matter
end


n = size(r,1);

if (nargin < 9),
  
  switch lsolver 
   case {1, 'infl-gmres'},
      %--flag is used to get rid off unnecessary screen outputs
      [z, flag]= gmres(@Kinfl, r, n, tol, itmax, [], [], [], lam, x); 
   case {2, 'cons-gmres'}
      [z, flag]= gmres(@Kcons, r, n, tol, itmax, [], [], [], lam, x); 
   case {3, 'rand-gmres'}
      [z, flag]= gmres(@Krand, r, n, tol, itmax, [], [], [], lam, x);     
   case {4, 'infl-bicgs'},
      [z]= tk_bicgstab(@Kinfl, r, tol, itmax, lam, x);
   case {5, 'cons-bicgs'}
      [z]= tk_bicgstab(@Kcons, r, tol, itmax, lam, x);
   case {6, 'rand-bicgs'}
      [z]= tk_bicgstab(@Krand, r, tol, itmax, lam, x);
   case {7, 'infl-minres'}, %works only if A and the perturbation are symmetric
      [z, flag]= minres(@Kinfl, r, tol, itmax, [], [], [], lam, x);
   case {8, 'consm-gmres'}
      [z, flag]= gmres(@Kconsm, r, n, tol, itmax, [], [], [], lam, x);      
   case {9, 'consm-bicgs'}
      [z]= tk_bicgstab(@Kconsm, r, tol, itmax, lam, x);     
   case {10, 'rpx-gmres'}
      global  beta_loc
      if abs(lam) > 1, beta_loc = lam;  else, beta_loc =1; end
      [z, flag]= gmres(@Krpx, r, n, tol, itmax, [], [], [], lam, x); 
   case {11, 'rpxa-gmres'}
      global  beta_loc
      if abs(lam) > 1, beta_loc = lam;  else, beta_loc =1; end
      [z, flag]= gmres(@Krpxa, r, n, tol, itmax, [], [], [], lam, x);       
   case {12, 'rpx-bicgs'},    
      global  beta_loc
      if abs(lam) > 1, beta_loc = lam;  else, beta_loc =1; end
      [z]= tk_bicgstab(@Krpx, r, tol, itmax, lam, x);     
   case {13, 'rpxa-bicgs'}
      global  beta_loc
      if abs(lam) > 1, beta_loc = lam;  else, beta_loc =1; end       
      [z]= tk_bicgstab(@Krpxa, r, tol, itmax, lam, x);         
   case {14, 'sym-gmres'}
      [z, flag]= gmres(@Ksym, r, n, tol, itmax, [], [], [], lam, x);          
   case {15, 'sym-bicgs'},
      [z]= tk_bicgstab(@Ksym, r, tol, itmax, lam, x);        
      
   otherwise
      error('no such lsolver type');
  end
  
else
  
  switch lsolver 
   case {1, 'infl-gmres'},
      [z, flag]= gmres(@KinflA, r, n, tol, itmax, [], [], [], lam, x, A); 
   case {2, 'cons-gmres'}
      [z, flag]= gmres(@KconsA, r, n, tol, itmax, [], [], [], lam, x, A); 
   case {3, 'rand-gmres'}
      [z, flag]= gmres(@KrandA, r, n, tol, itmax, [], [], [], lam, x, A);   
   case {4, 'infl-bicgs'},
      [z]= tk_bicgstab(@KinflA, r, tol, itmax, lam, x, A);
   case {5, 'cons-bicgs'}
      [z]= tk_bicgstab(@KconsA, r, tol, itmax, lam, x, A);
   case {6, 'rand-bicgs'}
      [z]= tk_bicgstab(@KrandA, r, tol, itmax, lam, x, A);
   case {7, 'infl-minres'}, %works only if A and the perturbation are symmetric
      [z, flag]= minres(@KinflA, r, tol, itmax, [], [], [], lam, x, A);
   case {8, 'consm-gmres'}
      [z, flag]= gmres(@KconsmA, r, n, tol, itmax, [], [], [], lam, x, A);      
   case {9, 'consm-bicgs'}
      [z]= tk_bicgstab(@KconsmA, r, tol, itmax, lam, x, A);           
   case {10, 'rpx-gmres'}
      [z, flag]= gmres(@KrpxA, r, n, tol, itmax, [], [], [], lam, x, A);   
   case {11, 'rpxa-gmres'}
      global  beta_loc
      if abs(lam) > 1, beta_loc = lam;  else, beta_loc =1; end
      [z, flag]= gmres(@KrpxaA, r, n, tol, itmax, [], [], [], lam, x, A);       
   case {12, 'rpx-bicgs'},    
      global  beta_loc
      if abs(lam) > 1, beta_loc = lam;  else, beta_loc =1; end
      [z]= tk_bicgstab(@KrpxA, r, tol, itmax, lam, x, A);     
   case {13, 'rpxa-bicgs'}
      global  beta_loc
      if abs(lam) > 1, beta_loc = lam;  else, beta_loc =1; end       
      [z]= tk_bicgstab(@KrpxaA, r, tol, itmax, lam, x, A);         
   case {14, 'sym-gmres'}
      [z, flag]= gmres(@KsymA, r, n, tol, itmax, [], [], [], lam, x, A);          
   case {15, 'sym-bicgs'},
      [z]= tk_bicgstab(@KsymA, r, tol, itmax, lam, x, A); 

   otherwise
      error('no such lsolver type');
  end
  
end

  nrmz=norm(z);
  if (nrmz == 0),
    %--- output lengthy but noticeable warnings ---
    fprintf('*** Warning: in modrqi.m, solver=%s, ||z||=%e\n', ...
	    lsolver, nrmz); 
    fprintf('*** This is caused by the solver used \n')
    fprintf('*** returned a zero vector instead of a \n'),
    fprintf('*** non-zero vector at the exit iteration step\n')
    fprintf('=== Now: replace solution z by a random vector\n')
    z = randn(size(x));
  end    

 
  
%========================================================================
function [y]=Kinfl(r, lam, x)
%
% compute  y =  (A-lamI + alpha*x*x')*r
%
% by sherman-morrison-woodbury formula, the above matrix is singular
% iff  1 + alpha* x'*(A-lamI)^{-1}*x = 0.
%
  global  alpha_loc 
  atmp = x'*r;
  
  %if 1 + alpha_loc*atmp == 0
  %  alpha_loc = 2*alpha_loc;
  %end

  w = user_Hx (r);  %get w = A*r;
  y = w - lam*r + alpha_loc*atmp*x;

  
  
%========================================================================
function [y]=Kcons(r, lam, x)
%
% compute  y =  (A- lamI - x*x'*(A+A'))*r  
%
  global A_operator  Matsymm 

  %Ax = r + lam*x;    %note r=Ax-lam x may not hold, r may be (I-VV')r
  %-for the symmetric case:
  %-there is no need to compute or pass-in A*x, we use x'*(A*r) term 
  %-directly instead of (x'*A')*r, since A*r always has to be computed
  %%Ax = user_Hx (x);  

  if (Matsymm),
    w = user_Hx (r); 
    atmp = x'*w;
    y = w - lam*r -2*atmp*x;
  else
    w = user_Hx (r);
    %-only implement the A isnumeric case
    %-the A*x can be passed from gdav.m to save the A'*r computation 
    %-(the trick is x'A'r=(Ax)'r, where Ax is computed in gdav.m;
    %- but this is not implemented here)
    atmp = x'*(w+A_operator'*r); 
    y = w - lam*r - x*atmp;
  end
  

%========================================================================
function [y]=Kconsm(r, lam, x)
%
% compute  y =  (A-lamI + alpha*x*x'*(A+A'))*r  
% simplified to   y =  (A-lamI + alpha*x*x'*A)*r
%
  global alpha_loc  %not equal to -2 in order to differ from Kcons.
  global A_operator  Matsymm 

  if (Matsymm | size(r,1) > 8000),
    w = user_Hx (r); 
    atmp = x'*w; 
    y = w - lam*r + alpha_loc*atmp*x;
    %y = w - lam*r  +  x*atmp;
  else
    w = user_Hx (r);
    atmp =x'*(w+A_operator'*r);  
    y = w - lam*r  +  alpha_loc* x*atmp;
    %y = w - lam*r  +  x*atmp;
  end
   
  
%========================================================================
function [y]=Krpx(r, lam, x)
%
% compute  y =  (A-lamI + alpha*x*(r+ beta x)')*r, where
%           
%    beta = lam  if |lam|>1; and beta = 1 if |lam|<=1.
%    (this is to deal with some ill conditioned matrix where the
%     large eigenvalues are computed, e.g., compute LR of 
%     the sheman3 with cond>6e+16)
%
  global  alpha_loc  beta_loc
  %if abs(lam) > 1, beta_loc = lam;  else, beta_loc =1; end
  
  w = user_Hx (r);  
  atmp = (r+ beta_loc*x)'*r;   %without reusing the A*r info in w
  %atmp = (r+ beta_loc*x)'*w;
  
  %y = w - lam*r + alpha_loc*atmp*x;
  y = w - lam*r -  atmp*x;
  
%========================================================================
function [y]=Krpxa(r, lam, x)
%
% compute  y =  (A-lamI + alpha*x*(r+ beta x)'A)*r, where
%           
%    beta = lam  if |lam|>1; and beta = 1 if |lam|<=1.
%    (this is to deal with some ill conditioned matrix where the
%     large eigenvalues are computed, e.g., compute LR of 
%     the sheman3 with cond>6e+16)
%
  global  alpha_loc  beta_loc
  
  w = user_Hx (r);  
  atmp = (r+ beta_loc*x)'*w;  %reusing the A*r info in w

  y = w - lam*r + alpha_loc*atmp*x;
  %y = w - lam*r - atmp*x;
  
  
%========================================================================
function [y]=Ksym(r, lam, x)
%
% compute  y =  (A - lamI -x*r'-r*x'- alpha*x*x')*r  
%
  global  alpha_loc  
  xdr = x'*r;
  atmp = r'*r + alpha_loc* xdr;
  
  w = user_Hx (r);  %get w = A*r;
  %--the following is inefficient 
  %%%  y = w - (lam + xdr)*r - (atmp)*x;
  %--taking out the rx' term makes it a lot better 
  %(then the equation is in form [ A -lam I + alpha x y'] t = -r)
  y = w - lam*r - (atmp)*x;
  
  
%========================================================================
function [y]=Krand(r, lam, x)
%
% compute  y =  (A-lamI + alpha*x*z')*r   where z=rand(n,1);
%
% by sherman-morrison-woodbury formula, the above matrix is singular
% iff  1 + alpha*z'*(A-lamI)^{-1}*x =0.
%
% moreover, an additional constraint for the vector z is that z should 
% not be parallel to r.  (this usually is true since z is random, 
% no need for a safeguard)
%
  global  alpha_loc 
  z = rand(size(r));

  %while (norm(z-r, inf)<1e-14 | norm(z+r, inf)<1e-14),
  %  z = rand(size(r));
  %end
  
  atmp = z'*r;
  w = user_Hx (r);  %get w = A*r;
  y = w - lam*r + alpha_loc*atmp*x;


  
  
%------------------------------------------------------------------------
% the following are mainly useful for A not in numeric form
%------------------------------------------------------------------------
  
%========================================================================
function [y]=KinflA(r, lam, x, A)
%
% compute  y =  (A-lamI + alpha*x*x')*r
%
% by sherman-morrison-woodbury formula, the above matrix is singular
% iff  1 + alpha* x'*(A-lamI)^{-1}*x =0.
%
  global  alpha_loc 
  atmp = x'*r;
  
  w = user_Hx (r, A);  %get w = A*r;
  y = w - lam*r + alpha_loc*atmp*x;

  
%========================================================================
function [y]=KconsA(r, lam, x, A)
%
% compute  y =  (A-lamI - 2*x*(A*x)')*r 
%
%  
  w = user_Hx (r, A);  %get w = A*r;
  atmp = w'*r;
  y = w - lam*r -2 *atmp*x;


%========================================================================
function [y]=KconsmA(r, lam, x)
%
% compute  y =  (A-lamI + alpha*x*(A*x)')*r  
%
  global alpha_loc
  
  w = user_Hx (r, A);  %get w = A*r;
  atm = w'*r;
  y = w - lam*r + alpha_loc*atmp*x;  
  
%========================================================================
function [y]=KrpxA(r, lam, x)
%
% compute  y =  (A-lamI + alpha*x*(r+x)')*r  
%
  global  alpha_loc   beta_loc
  
  atmp = (r+ beta_loc*x)'*r;
  
  w = user_Hx (r, A);  %get w = A*r;
  y = w - lam*r + alpha_loc*atmp*x;

%========================================================================
function [y]=KrpxaA(r, lam, x)
%
% compute  y =  (A-lamI + alpha*x*(r+ beta x)'A)*r, where
%           
%    beta = lam  if |lam|>1; and beta = 1 if |lam|<=1.
%    (this is to deal with some ill conditioned matrix where the
%     large eigenvalues are computed, e.g., compute LR of 
%     the sheman3 with cond>6e+16)
%
  global  alpha_loc  beta_loc
  
  w = user_Hx (r, A);  
  atmp = (r+ beta_loc*x)'*w;  %reusing the A*r info in w

  %y = w - lam*r + alpha_loc*atmp*x;
  y = w - lam*r - atmp*x;
  
  
%========================================================================
function [y]=KsymA(r, lam, x)
%
% compute  y =  (A - lamI -x*r'-r*x'- alpha*x*x')*r  
%
  global  alpha_loc  
  xr = x'*r;
  atmp = r'*r + alpha_loc* xr;
  
  w = user_Hx (r, A);  %get w = A*r;
  y = w - (lam + xr)*r - (atmp)*x;

  
%========================================================================
function [y]=KrandA(r, lam, x, A)
%
% compute  y =  (A-lamI + alpha*x*z')*r   where z=randn(n,1);
%
% by sherman-morrison-woodbury formula, the above matrix is singular
% iff  1 + alpha*z'*(A-lamI)^{-1}*x =0,
%
% moreover, an additional constraint for the vector z is that z is .
% not parallel to r. (this should be true since z is random)
%
  global alpha_loc
  z = randn(size(r));
  atmp = z'*r;
  
  w = user_Hx (r, A);  %get w = A*r;
  y = w - lam*r + alpha_loc*atmp*x;


