function [x, error, total_iters]=tk_bicgstab(atv, b, tol, kmax, lam, yy, eigv, A, x0);
%
% Bi-CGSTAB solver for linear systems
%
% C. T. Kelley, December 16, 1994
%
% modified by ykzhou to fit numerical tests interface.
%
%---------------------------------------------------------------
% This code comes with no guarantee or warranty of any kind.
%
% function [x, error, total_iters] = bicgstab(x0, b, atv, params)
%
%
% Input:        x0=initial iterate
%               b=right hand side
%               atv, a matrix-vector product routine
%			atv must return Ax when x is input
%			the format for atv is
%                       function ax = atv(x)
%               params = two dimensional vector to control iteration
%                        params(1) = relative residual reduction factor
%                        params(2) = max number of iterations
%
% Output:       x=solution
%               error = vector of iteration residual norms
%               total_iters = number of iterations
%

%
% initialization
%
n=length(b); 
if (nargin < 9), x0=zeros(n,1); end
errtol = tol*norm(b); error=[]; x=x0;
rho=zeros(kmax+1,1);
%
if norm(x)~=0
  if (nargin >=8),
    r=b-feval(atv, x, lam, yy, eigv, A);
  elseif (nargin >=7),
    r=b-feval(atv, x, lam, yy, eigv);
  elseif (nargin >=6),
    r=b-feval(atv, x, lam, yy);
  elseif (nargin >=5),
    r=b-feval(atv, x, lam);
  else
    r=b-feval(atv, x);
  end
else
  r=b;
end
error=[];
%
hatr0=r;
k=0; rho(1)=1; alpha=1; omega=1;
v=zeros(n,1); p=zeros(n,1); rho(2)=hatr0'*r;
zeta=norm(r); error=[error,zeta];
%
% Bi-CGSTAB iteration
%
while((zeta > errtol) & (k < kmax))
    k=k+1;
    if omega==0
       error('Bi-CGSTAB breakdown, omega=0');
    end
    beta=(rho(k+1)/rho(k))*(alpha/omega);
    p=r+beta*(p - omega*v);
    if (nargin >=8),
      v=feval(atv, p, lam, yy, eigv, A);
    elseif (nargin >=7),
      v=feval(atv, p, lam, yy, eigv);
    elseif (nargin >=6),    
      v=feval(atv, p, lam, yy);
    elseif (nargin >=5),
      v=feval(atv, p, lam);
    else
       v=feval(atv,p);
    end

    tau=hatr0'*v;
    if tau==0
        error('Bi-CGSTAB breakdown, tau=0');
    end 
    alpha=rho(k+1)/tau;
    s=r-alpha*v; 

    if (nargin >=8),
      t=feval(atv, s, lam, yy, eigv, A);
    elseif (nargin >=7),
      t=feval(atv, s, lam, yy, eigv);
    elseif (nargin >=6),  
      t=feval(atv, s, lam, yy);
    elseif (nargin >=5),
      t=feval(atv, s, lam);
    else
      t=feval(atv, s);
    end
    
    tau=t'*t;
    if tau==0
       error('Bi-CGSTAB breakdown, t=0');
    end
    omega=t'*s/tau; 
    rho(k+2)=-omega*(hatr0'*t);
    x=x+alpha*p+omega*s;
    r=s-omega*t;
    zeta=norm(r);
    total_iters=k;
    error=[error, zeta];
end
%fprintf('--bicgstab: norm(res)=%e,  k=%i,  itmax=%i\n',zeta, k, kmax);